/*-----------------------------------------------------------------
  All code is provided for EDUCATIONAL purposes ONLY. There is no
  warranty and no support.

  Released under the MIT License  - see
  http://www.opensource.org/licenses/mit-license.php

  Outdoor Thermometer Copyright (c) 2015 by Jack Botner

  main.c

  This program implements a thermometer using a 4 digit LED display.
  The sensor, an LM19, may be located outdoors, connected with a 3
  conductor cable. The unit is powered by a 5 volt (USB type) wall
  wart, and is on all the time, so the nice bright display can be
  read day or night. There are two buttons to enter a correction 
  factor to make the thermometer more accurate.

  Processor: ATMega48
      Clock: Internal, 8.0 MHz
-----------------------------------------------------------------*/
#define F_CPU 8000000UL			// 8 MHz
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <util/delay_basic.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "led.h"

#define VREF_MV			3300		// Vref (measured) (mV)
#define NR_AVERAGES     10

#define EVENT_NONE			0
#define EVENT_DISPLAY_DATA	1
#define EVENT_LED_REFRESH	2
#define EVENT_DEBOUNCE		3
#define EVENT_BUTTON1_PRESS	4
#define EVENT_BUTTON2_PRESS	5

// Global variables

static volatile uint8_t ucLed[4] = { 9, 9, 9, 9 };

static volatile uint16_t uiLastAdc = 0;
static volatile int16_t  iTempOffset = 0;

static volatile uint8_t ucEventFlags = EVENT_NONE;

// Internal Functions

static int averager( int iTemp );
static int convert_adc( uint16_t uiAdc );
static void format_display( int iTemp );
static void system_init(void);

/*-----------------------------------------------------------------
  main()
-----------------------------------------------------------------*/
int main( void )
{
  static volatile uint8_t ucIndex=0;		// digit selection index
  static volatile uint8_t ucBtn1Press=0;	// button1 press counter
  static volatile uint8_t ucBtn2Press=0;	// button2 press counter
  uint16_t	uiAdc;
  uint8_t	ucNr;
  int       iTemp;

  system_init();                		// initialize system

  // Get previous temperature offset from EEprom (10ths degrees)
  iTempOffset = eeprom_read_word( 0 );
  if ( iTempOffset > 199 || iTempOffset < -199 )	// probably not initialized
  {
    iTempOffset = 0;
	eeprom_write_word( 0, iTempOffset );
  }

  while (1)
  {
	if ( test_bit( ucEventFlags, EVENT_LED_REFRESH ) )
	{
	  ucNr = ucLed[ucIndex];
	  ledOutputNumber( ucNr, ucIndex );

	  ledSelectDigit( ucIndex );

	  ++ucIndex;
	  if ( ucIndex > 3 )
	    ucIndex = 0;  

	  clear_bit( ucEventFlags, EVENT_LED_REFRESH );
	}

	if ( test_bit( ucEventFlags, EVENT_DEBOUNCE ) )
	{
	  if ( !test_bit( PINB, PB0 ) )
	  {
	    ++ucBtn1Press;
		if ( ucBtn1Press == 8 )		// ~20ms
		{
		  set_bit( ucEventFlags, EVENT_BUTTON1_PRESS );
		}
	  }
	  else
	    ucBtn1Press = 0;

	  if ( !test_bit( PINB, PB1 ) )
	  {
		  ++ucBtn2Press;
		  if ( ucBtn2Press == 8 )		// ~20ms
		  {
			  set_bit( ucEventFlags, EVENT_BUTTON2_PRESS );
		  }
	  }
	  else
	    ucBtn2Press = 0;

	  clear_bit( ucEventFlags, EVENT_DEBOUNCE );
	}

	if ( test_bit( ucEventFlags, EVENT_DISPLAY_DATA ) )
	{
	  // uiAdc = digital_filter( DIG_GET, 0 );
      uiAdc = uiLastAdc;
	  iTemp = convert_adc( uiAdc );
	  iTemp = averager( iTemp );
	  format_display( iTemp );
	  clear_bit( ucEventFlags, EVENT_DISPLAY_DATA );
	}

	if ( test_bit( ucEventFlags, EVENT_BUTTON1_PRESS ) )
	{
	  // In response to a button1 press we want to increment the
	  // temperature offset reading by 1/10th degree.

      if ( iTempOffset < 200 )
	  {
	    ++iTempOffset;
		eeprom_write_word( 0, iTempOffset );
        set_bit( ucEventFlags, EVENT_DISPLAY_DATA );
	  }
	  
	  clear_bit( ucEventFlags, EVENT_BUTTON1_PRESS );
	}

    if ( test_bit( ucEventFlags, EVENT_BUTTON2_PRESS ) )
    {
	  // In response to a button1 press we want to decrement the
	  // temperature offset reading by 1/10th degree.

	  if ( iTempOffset > -200 )
	  {
		--iTempOffset;
		eeprom_write_word( 0, iTempOffset );
        set_bit( ucEventFlags, EVENT_DISPLAY_DATA );
	  }
	
	  clear_bit( ucEventFlags, EVENT_BUTTON2_PRESS );
    }
  }
}

/*-----------------------------------------------------------------
  ISR(TIMER2_OVF_vect)

  Timer 2 overflow interrupt handler

  Since the clock is running at 8.0 mHz, the counter increments
  every 125 ns. The counter overflows after 256 counts, so an
  interrupt would occur every 32.0 us. We specify a prescaler of
  128, so the interrupt actually occurs every 4.096 ms.
  This is a good interval to use to switch LED digits.
-----------------------------------------------------------------*/
ISR(TIMER2_OVF_vect)
{
  static volatile uint8_t  ucCtr1=0;		// interrupt counter
  static volatile uint16_t uiCtr2=0;		// interrupt counter

  // Signal an LED refresh event
  set_bit( ucEventFlags, EVENT_LED_REFRESH );

  // Signal a button debounce event
  set_bit( ucEventFlags, EVENT_DEBOUNCE );

  // Every 125 interrupts (~500 ms) start an ADC conversion
  // FYI, the adc clock is running at 125kHz (8 us) and
  // a conversion takes 13 clock cycles (104 us).
  ++ucCtr1;
  if ( ucCtr1 == 125 )
  {
  	set_bit( ADCSRA, ADSC );	// start an ADC conversion
	ucCtr1 = 0;
  }

  // Every 1250 interrupts (~5 seconds) post an event
  // to update the display
  ++uiCtr2;
  if ( uiCtr2 == 1250 )
  {
  	set_bit( ucEventFlags, EVENT_DISPLAY_DATA );
	uiCtr2 = 0;
  }
}

/*-----------------------------------------------------------------
  ISR( ADC_vect ) - ADC conversion complete interrupt handler

  ADC0 is monitoring the voltage across the unknown resistor.
-----------------------------------------------------------------*/
ISR( ADC_vect )
{
  uint16_t	uiAdc;

  // ADCL must be read before ADCH.
  uiAdc = (uint16_t) ADCL | ( (uint16_t) ADCH << 8 );

  //digital_filter( DIG_ADD, uiAdc );
  uiLastAdc = uiAdc;
}

/*-----------------------------------------------------------------
  system_init()

  PORTB: PB0 (input) calibrate pushbutton 1
         PB1 (input) calibrate pushbutton 2
         PB2 (input)  not used
         PB3 (output) MOSI ISP
         PB4 (output) MISO ISP
         PB5 (output) SCK ISP
         PB6 (input)  not used
         PB7 (input)  not used
		 
  PORTC: PC0 (input)  ADC0 LM34
         PC1 (input)  not used
         PC2 (output) LED Digit 1 select
         PC3 (output) LED Digit 2 select
         PC4 (output) LED Digit 3 select
         PC5 (output) LED Digit 4 select
         PC6 (input) !Reset

  PORTD: PD0 (output) LED A segment
         PD1 (output) LED B segment
         PD2 (output) LED C segment
         PD3 (output) LED D segment
         PD4 (output) LED E segment
         PD5 (output) LED F segment
         PD6 (output) LED G segment
         PD7 (output) LED decimal point
-----------------------------------------------------------------*/
static void system_init()
{
  // The fuses are set to provide an internal clock, running at
  // 8.0 MHz with a prescaler of 8. Change the prescaler from
  // 8 to 1. The procedure requires two steps as per the device
  // specifications.
  CLKPR = 0b10000000;       // set clock prescaler change enable
  CLKPR = 0b00000000;       // set prescaler to 1

  // Set up the I/O ports directions, pull-ups
  DDRB  = 0b00111000;       // PortB I/O
  PORTB = 0b11000111;       // initialize PortB, pullups

  DDRC  = 0b00111100;       // PortC I/O
  PORTC = 0b00111110;       // initialize PortC7

  DDRD  = 0b11111111;       // PortD0 I/O
  PORTD = 0b11111111;       // initialize PortD7

  // Set up Timer/Counter2 to produce an interrupt every 4 ms.
  TCCR2A = 0b00000000;		// normal operation
  TCCR2B = 0b00000101;		// use prescaler=128
  TCNT2 = 0;
  TIMSK2 = 0b00000001;		// enable interrupt on overflow	

  // Set up PC0/ADC0 as analog input, external Aref
  ADMUX  = 0b00000000;		// external Aref; result right adjusted; select ADC0
  ADCSRA = 0b10001110;		// enable ADC; Enable interrupt; prescaler = 64
  ADCSRB = 0b00000000;		// auto trigger source = none
  DIDR0  = 0b00000001;		// disable PC0 digital input buffer

  PRR = 0b10000110;			// power save: no TWI, no SPI, no USART0
  sei();					// enable global interrupts
}

/*-----------------------------------------------------------------
  convert_adc(...)

  Input: current A/D reading

  Output: temperature in tenths degrees C
-----------------------------------------------------------------*/
static int convert_adc( uint16_t uiAdc )
{
  uint16_t	uiMV;
  int       iTemp;
  double    dVo, dTerm1, dTerm2, dTerm3;

  // Convert the A/D output to millivolts.
  // A/D is in the range 0..1023
  // Vref is 1100 mV

  uiMV = (uint16_t) ( ( (uint32_t) uiAdc * VREF_MV ) / 1024 );

  // Convert milllivolts to degrees C
  // Gotta use floating point for this
  /*
                         -------------------------------
                        /                 1.8639 - Vo
      T = -1481.96 +   / 2.1962x10E6 + --------------
                     \/                   3.88x10E-6
  */
  // Instead of one humongous calculation I like to break it down into parts

  dVo = (double) uiMV / 1000.0;		// Vo, volts
  dTerm1 = 1.8639 - dVo;
  dTerm2 = dTerm1 / 3.88E-6;
  dTerm3 = 2.1962E6 + dTerm2;
  dTerm1 = sqrt( dTerm3 );
  dTerm2 = -1481.96 + dTerm1;		// degrees C
  dTerm3 = dTerm2 * 10.0;           // tenths degrees C

  iTemp = (int) round( dTerm3 ) + iTempOffset;
  return iTemp;
}

/*-----------------------------------------------------------------
  format_display(...)

  Input: temperature, 10ths degrees C

  Output: ucLed[0..3]
-----------------------------------------------------------------*/
static void format_display( int iTemp )
{
  uint16_t	uiNr, uiTemp;
  uint8_t	ucNonZeroIndex=0, ucNegative=0;

  if ( iTemp < 0 ) ucNegative = 1;  
  uiTemp = (unsigned) abs( iTemp );

  // Convert result into 4 digits in ucLed

  uiNr = uiTemp / 1000;
  ucLed[0] = (uint8_t) uiNr;
  if ( ucLed[0] )
    ucNonZeroIndex = 1;
  else
	ucLed[0] = 0x0f;			// suppress leading zero
  uiTemp -= ( uiNr * 1000 );

  uiNr = uiTemp / 100;
  ucLed[1] = (uint8_t) uiNr;
  if ( !ucNonZeroIndex )
  {
	if ( ucLed[1] )
      ucNonZeroIndex = 1;
	else
	  ucLed[1] = 0x0f;			// suppress leading zero
  }

  if ( ucNegative )
  {
	if ( uiTemp > 99 )
	ucLed[0] = LED_VALUE_MINUS;
	else
	ucLed[1] = LED_VALUE_MINUS;
  }

  uiTemp -= ( uiNr * 100 );

  uiNr = uiTemp / 10;
  ucLed[2] = (uint8_t) uiNr;
#if 0
  if ( !ucNonZeroIndex )
  {
	if ( ucLed[2] )
      ucNonZeroIndex = 1;
	else
	  ucLed[2] = 0x0f;			// suppress leading zero
  }
#endif  
  uiTemp -= ( uiNr * 10 );

  ucLed[3] = (uint8_t) uiTemp;
}

/*-----------------------------------------------------------------
  averager(...)

  Generate the average of NR_AVERAGES samples. I do this to
  correct a problem of the display hunting between two close
  readings.
  
  Input: temperature, 10ths degrees C

  Returns: temperature, 10ths degrees C
-----------------------------------------------------------------*/
static int averager( int iTemp )
{
  static volatile int iAverages[NR_AVERAGES];
  static volatile uint8_t ucFirst=1;
  int     iTmp, ii;

  if ( ucFirst )
  {
	// Initialize the array with the 1st value
	for ( ii=0; ii<NR_AVERAGES; ++ii )
	   iAverages[ii] = iTemp;
	ucFirst = 0;   
  }

  // Move array up one position
  for ( ii=0; ii<NR_AVERAGES-1; ++ii )
    iAverages[ii] = iAverages[ii+1];
  
  // Insert current vale at end of array
  iAverages[NR_AVERAGES-1] = iTemp;
  
  // Average the array
  for ( ii=0, iTmp=0; ii<NR_AVERAGES-1; ++ii )
    iTmp += iAverages[ii];
  
  iTmp /= NR_AVERAGES;
  return iTmp;
}
